"""
# Usage
- Set `CONNECT_WITH_STUDY_SERVER` in overcooked_gui.py to True.
- Run this script. Copy the manager id that is printed
- Run the game_server.py script with the manager id copied from the terminal
```
python game_server.py --manager_ids COPIED_UUID
```
- Run 2 overcooked_gui.py scripts in different terminals. For more players change `NUMBER_PLAYER_PER_ENV` and start more guis.

The environment starts when all players connected.
"""

import argparse
import asyncio
import logging
from pathlib import Path
from typing import Tuple, TypedDict

import requests
import uvicorn
import yaml
from fastapi import FastAPI

from overcooked_simulator import ROOT_DIR
from overcooked_simulator.game_server import CreateEnvironmentConfig
from overcooked_simulator.server_results import PlayerInfo
from overcooked_simulator.utils import (
    url_and_port_arguments,
    add_list_of_manager_ids_arguments,
)

NUMBER_PLAYER_PER_ENV = 2

log = logging.getLogger(__name__)

app = FastAPI()

game_server_url = "localhost:8000"
server_manager_id = None


# @app.get("/")
# async def root(response_class=HTMLResponse):
#     return """
#     <html>
#         <head>
#             <title>Overcooked Game</title>
#         </head>
#         <body>
#             <h1>Start Game!</h1>
#             <button type="button">Click Me!</button>
#         </body>
#     </html>
#     """


HARDCODED_MANAGER_ID = "1234"


running_tutorials: dict[str, Tuple[int, dict[str, PlayerInfo], list[str]]] = {}


class LevelConfig(TypedDict):
    name: str
    config_path: str
    layout_path: str
    item_info_path: str


class StudyConfig(TypedDict):
    levels: list[LevelConfig]
    num_players: int


class StudyState:
    def __init__(self, study_config_path: str | Path):
        with open(study_config_path, "r") as file:
            env_config_f = file.read()

        self.study_config: StudyConfig = yaml.load(
            str(env_config_f), Loader=yaml.SafeLoader
        )
        self.levels: list[LevelConfig] = self.study_config["levels"]
        self.current_level_idx: int = 0

        self.participant_id_to_player_info = {}
        self.player_ids = {}
        self.num_connected_players: int = 0

        self.current_running_env = None
        self.next_level_env = None
        self.players_done = {}

    @property
    def study_done(self):
        return self.current_level_idx >= len(self.levels)

    @property
    def is_full(self):
        return (
            len(self.participant_id_to_player_info) == self.study_config["num_players"]
        )

    def create_env(self, level):
        with open(ROOT_DIR / "game_content" / level["item_info_path"], "r") as file:
            item_info = file.read()
        with open(
            ROOT_DIR / "game_content" / "layouts" / level["layout_path"], "r"
        ) as file:
            layout = file.read()
        with open(ROOT_DIR / "game_content" / level["config_path"], "r") as file:
            environment_config = file.read()

        creation_json = CreateEnvironmentConfig(
            manager_id=server_manager_id,
            number_players=self.study_config["num_players"],
            environment_settings={"all_player_can_pause_game": False},
            item_info_config=item_info,
            environment_config=environment_config,
            layout_config=layout,
            seed=1234567890,
        ).model_dump(mode="json")

        env_info = requests.post(
            game_server_url + "/manage/create_env/", json=creation_json
        )

        if env_info.status_code == 403:
            raise ValueError(f"Forbidden Request: {env_info.json()['detail']}")
        env_info = env_info.json()

        return env_info

    def start(self):
        level = self.levels[self.current_level_idx]
        print("CREATING LEVEL:", level["name"])
        self.current_running_env = self.create_env(level)
        # if len(self.levels) > 1:
        # next_level = self.levels[self.current_level_idx]
        # print("CREATING NEXT LEVEL:", next_level["name"])
        # self.next_level_env = self.create_env(next_level)

    def next_level(self):
        requests.post(
            f"{game_server_url}/manage/stop_env/",
            json={
                "manager_id": HARDCODED_MANAGER_ID,
                "env_id": self.current_running_env["env_id"],
                "reason": "Next level",
            },
        )

        self.current_level_idx += 1
        if not self.study_done:
            level = self.levels[self.current_level_idx]
            print("CREATING LEVEL:", level["name"])
            self.current_running_env = self.create_env(level)
            for participant_id, player_id in self.player_ids.items():
                player_id = self.player_ids[participant_id]
                self.participant_id_to_player_info[
                    participant_id
                ] = self.current_running_env["player_info"][player_id]

            for key in self.players_done:
                self.players_done[key] = False
            return False
        else:
            return True

    def add_participant(self, participant_id: str):
        player_name = str(self.num_connected_players)
        player_info = self.current_running_env["player_info"][player_name]
        self.participant_id_to_player_info[participant_id] = player_info
        self.player_ids[participant_id] = player_info["player_id"]
        self.num_connected_players += 1
        return player_info

    def player_finished_level(self, participant_id):
        self.players_done[participant_id] = True
        level_done = all(self.players_done.values())
        if level_done:
            self.next_level()
        return {"study_finished": self.study_done}

    def get_connection(self, participant_id: str):
        player_info = self.participant_id_to_player_info[participant_id]
        return player_info


class StudyManager:
    def __init__(self):
        self.running_studies: list[StudyState] = []

        self.participant_id_to_study_map: dict[str, StudyState] = {}
        self.running_envs: dict[str, Tuple[int, dict[str, PlayerInfo], list[str]]] = {}
        self.current_free_envs = []

    def create_study(self):
        study = StudyState(ROOT_DIR / "game_content" / "study_config.yaml")
        study.start()
        self.running_studies.append(study)

    def add_participant(self, participant_id):
        player_info = None
        if all([s.is_full for s in self.running_studies]):
            self.create_study()

        for study in self.running_studies:
            if not study.is_full:
                player_info = study.add_participant(participant_id)
                self.participant_id_to_study_map[participant_id] = study
        return player_info

    def player_finished_level(self, participant_id: str):
        assigned_study = self.participant_id_to_study_map[participant_id]
        return assigned_study.player_finished_level(participant_id)

    def get_participant_game_connection(self, participant_id: str):
        assigned_study = self.participant_id_to_study_map[participant_id]
        return assigned_study.participant_id_to_player_info[participant_id]


study_manager = StudyManager()


@app.post("/start_study/{participant_id}")
async def start_study(participant_id: str):
    player_info = study_manager.add_participant(participant_id)
    print()
    return player_info


@app.post("/level_done/{participant_id}")
async def next_level(participant_id: str):
    is_there_next_level = study_manager.player_finished_level(participant_id)
    return is_there_next_level


@app.post("/get_game_connection/{participant_id}")
async def finished_game(participant_id: str):
    return study_manager.get_participant_game_connection(participant_id)


@app.post("/finished_game/{participant_id}")
async def finished_game(participant_id: str):
    print(f"{participant_id} finished game.")
    return None


@app.post("/connect_to_tutorial/{participant_id}")
async def want_to_play_tutorial(participant_id: str):
    environment_config_path = (
        ROOT_DIR / "game_content" / "tutorial" / "tutorial_env_config.yaml"
    )
    layout_path = ROOT_DIR / "game_content" / "tutorial" / "tutorial.layout"
    item_info_path = ROOT_DIR / "game_content" / "item_info.yaml"
    with open(item_info_path, "r") as file:
        item_info = file.read()
    with open(layout_path, "r") as file:
        layout = file.read()
    with open(environment_config_path, "r") as file:
        environment_config = file.read()
    creation_json = CreateEnvironmentConfig(
        manager_id=server_manager_id,
        number_players=1,
        environment_settings={"all_player_can_pause_game": False},
        item_info_config=item_info,
        environment_config=environment_config,
        layout_config=layout,
        seed=1234567890,
    ).model_dump(mode="json")
    # todo async
    print("CREATING TUTORIAL ENVIRONMENT")
    env_info = requests.post(
        game_server_url + "/manage/create_env/", json=creation_json
    )

    if env_info.status_code == 403:
        raise ValueError(f"Forbidden Request: {env_info.json()['detail']}")
    env_info = env_info.json()
    running_tutorials[participant_id] = env_info
    return env_info["player_info"]["0"]


@app.post("/disconnect_from_tutorial/{participant_id}")
async def want_to_play_tutorial(participant_id: str):
    requests.post(
        f"{game_server_url}/manage/stop_env/",
        json={
            "manager_id": HARDCODED_MANAGER_ID,
            "env_id": running_tutorials[participant_id]["env_id"],
            "reason": "Finished tutorial",
        },
    )


def main(host, port, game_server_url_, manager_ids):
    global game_server_url, server_manager_id

    manager_ids = ["1234"]

    game_server_url = "http://" + game_server_url_
    server_manager_id = manager_ids[0]
    print(f"Use {server_manager_id=} for {game_server_url=}")
    loop = asyncio.new_event_loop()
    config = uvicorn.Config(app, host=host, port=port, loop=loop)
    server = uvicorn.Server(config)
    loop.run_until_complete(server.serve())


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="Overcooked Simulator Study Server",
        description="Study Server: Match Making, client pre and post managing.",
        epilog="For further information, see https://scs.pages.ub.uni-bielefeld.de/cocosy/overcooked-simulator/overcooked_simulator.html",
    )
    url_and_port_arguments(parser=parser, server_name="Study Server", default_port=8080)
    add_list_of_manager_ids_arguments(parser=parser)
    args = parser.parse_args()
    main(
        args.url,
        args.port,
        game_server_url_="localhost:8000",
        manager_ids=args.manager_ids,
    )
